import time
from typing import Dict, Optional

import jax.numpy as jnp
from jax import device_put


def encode_inputs(*, inputs: jnp.ndarray, encoder: jnp.ndarray) -> jnp.ndarray:
    """
    Encodes input data using the provided encoder matrix.

    Parameters:
      inputs: jax array, shape (n_samples, input_dim).
      encoder: jax array, shape (input_dim, reduced_dim).

    Returns:
      jax array, shape (n_samples, reduced_dim), with encoded inputs.
    """
    return jnp.einsum("nx,xr->nr", inputs, encoder)


def encode_outputs(*, outputs: jnp.ndarray, encoder: jnp.ndarray) -> jnp.ndarray:
    """
    Encodes output data using the provided encoder matrix.

    Parameters:
      outputs: jax array, shape (n_samples, output_dim).
      encoder: jax array, shape (output_dim, reduced_dim).

    Returns:
      jax array, shape (n_samples, reduced_dim), with encoded outputs.
    """
    return jnp.einsum("no,or->nr", outputs, encoder)


def encode_Jacobians(
    *,
    jacobians: jnp.ndarray,
    input_decoder: Optional[jnp.ndarray] = None,
    output_encoder: Optional[jnp.ndarray] = None,
    batched: bool = False,
    batch_size: int = 10,
) -> jnp.ndarray:
    """
    Reduces Jacobians using input decoder and/or output encoder matrices.

    Parameters:
      jacobians: jax array, shape (n_samples, output_dim, input_dim).
      input_decoder: Optional; jax array of shape
        (input_dim, reduced_in_dim).
      output_encoder: Optional; jax array of shape
        (output_dim, reduced_out_dim).
      batched: Process reduction in batches (default: False).
        If batched, the jacobians may reside on the cpu a priori.
        They will be moved to gpu to be reduced.
      batch_size: Batch size when batched is True (default: 50).

    Returns:
      jax array with reduced dimensions:
        - If both input_decoder and output_encoder provided:
            (n_samples, reduced_out_dim, reduced_in_dim)
        - If only output_encoder provided:
            (n_samples, reduced_out_dim, input_dim)
        - If only input_decoder provided:
            (n_samples, output_dim, reduced_in_dim)
        - If neither provided, returns original jacobians.
    """

    def reduce_batch(jacs: jnp.ndarray) -> jnp.ndarray:
        if output_encoder is not None and input_decoder is not None:
            return jnp.einsum("ol,nox,xr->nlr", output_encoder, jacs, input_decoder)
        elif output_encoder is not None:
            return jnp.einsum("ol,nox->nlx", output_encoder, jacs)
        elif input_decoder is not None:
            return jnp.einsum("nox,xr->nor", jacs, input_decoder)
        else:
            return jacs

    if batched:
        total_len = jacobians.shape[0]
        reduced_batches = []
        for start in range(0, total_len, batch_size):
            end = min(start + batch_size, total_len)
            batch = device_put(jacobians[start:end])
            reduced_batches.append(reduce_batch(batch))
        return jnp.concatenate(reduced_batches, axis=0)
    else:
        return reduce_batch(jacobians)


def encode_input_output_Jacobian_data(
    *,
    inputs: jnp.ndarray,
    outputs: jnp.ndarray,
    jacobians: Optional[jnp.ndarray] = None,
    input_encoder: Optional[jnp.ndarray] = None,
    output_encoder: Optional[jnp.ndarray] = None,
    input_decoder: Optional[jnp.ndarray] = None,
    batched: bool = False,
    batch_size: int = 50,
) -> Dict[str, jnp.ndarray]:
    """
    Encodes inputs, outputs, and reduces Jacobians using the provided matrices.

    Parameters:
      inputs: jax array, shape (n_samples, input_dim).
      outputs: jax array, shape (n_samples, output_dim).
      jacobians:  Optional; jax array, shape (n_samples, output_dim, input_dim).
      input_encoder: Optional; for encoding inputs, shape
        (input_dim, reduced_dim).
      output_encoder: Optional; for encoding outputs, shape
        (output_dim, reduced_dim).
      input_decoder: Optional; for reducing jacobian input dim, shape
        (input_dim, reduced_in_dim).
      batched: Process jacobians in batches (default: False).
      batch_size: Batch size if batched is True.

    Returns:
      Dict with keys:
        "encoded_inputs": encoded inputs (or original if not provided).
        "encoded_outputs": encoded outputs (or original if not provided).
        "reduced_Jacobians": reduced jacobians (or original if not reduced).
    """
    print("Placing inputs/input_encoder on GPU if not already")
    inputs = device_put(inputs) if inputs is not None else None
    input_encoder = device_put(input_encoder) if input_encoder is not None else None

    start = time. perf_counter()
    encoded_inputs = encode_inputs(inputs=inputs, encoder=input_encoder) if input_encoder is not None else inputs
    input_encoding_time = time. perf_counter() - start

    print("Placing outputs/output_encoder on GPU if not already")
    outputs = device_put(outputs) if outputs is not None else None
    output_encoder = device_put(output_encoder) if output_encoder is not None else None

    start = time. perf_counter()
    encoded_outputs = encode_outputs(outputs=outputs, encoder=output_encoder) if output_encoder is not None else outputs
    output_encoding_time = time. perf_counter() - start
    encoded_dict = {
        "encoded_inputs": encoded_inputs,
        "encoded_outputs": encoded_outputs,
        "input_encoding_time": input_encoding_time,
        "output_encoding_time": output_encoding_time,
    }
    if jacobians is not None:

        input_decoder = device_put(input_decoder) if input_decoder is not None else None
        start = time. perf_counter()
        encoded_Jacobians = encode_Jacobians(
            jacobians=jacobians,
            input_decoder=input_decoder,
            output_encoder=output_encoder,
            batched=batched,
            batch_size=batch_size,
        )
        jacobian_encoding_time = time. perf_counter() - start

        encoded_dict["encoded_Jacobians"] = encoded_Jacobians
        encoded_dict["Jacobian_encoding_time"] = jacobian_encoding_time
    return encoded_dict



"""
BayesFlux: Bayesian Fast Linear algebra sUbspace eXtraction in JAX.
"""

from ._hippylib_utils import (
    hippylib_sampler_from_model,
    multiprocess_generate_hippylib_full_Jacobian_data,
    multiprocess_generate_hippylib_output_data,
    multiprocess_generate_reduced_hippylib_training_data,
)
from .generation import GaussianInputOuputAndDerivativesSampler, InputOuputAndDerivativesSampler



import time
from typing import Any, Dict

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
from jax.scipy.linalg import solve_triangular
from randlax import double_pass_randomized_eigh, double_pass_randomized_gen_eigh


@jax.jit
def __JCJtransp(J: jnp.ndarray, L: jnp.ndarray) -> jnp.ndarray:
    # Solve L Y = J.T for Y.
    Y = solve_triangular(L, J.T, lower=True)
    # Then, J C^-1 J.T = Y.T @ Y
    return Y.T @ Y


def average_JCJtranspose(J_samples: jnp.ndarray, prior_precision: jnp.ndarray) -> jnp.ndarray:
    L = jnp.linalg.cholesky(prior_precision)
    # Average over the batch dimension (axis=0)
    return jnp.mean(jax.vmap(lambda J: __JCJtransp(J, L))(J_samples), axis=0)


def average_Jtranspose_sigmainv_J_chunked(J, noise_precision, chunk_size=25):
    n, a, b = J.shape  # i.e. n=1000
    # n_chunks: i.e. 1000//25 = 40,
    Jc = J.reshape((n // chunk_size, chunk_size, a, b))

    def body(carry, block):
        return carry + jnp.tensordot(block, block, axes=((0, 1), (0, 1))), None

    init = jnp.zeros((b, b), J.dtype)
    A_sum, _ = jax.lax.scan(body, init, Jc)
    return (noise_precision / n) * A_sum


def average_JCoversigmaJtranpose_chunked(J_samples, prior_covariance, noise_precision, chunk_size=100):
    n, a, b = J_samples.shape
    assert n % chunk_size == 0, "n must be divisible by chunk_size"
    # reshape into (n_chunks, chunk_size, a, b)
    n_chunks = n // chunk_size
    Jc = J_samples.reshape(n_chunks, chunk_size, a, b)

    def scan_fn(acc, block):
        # block has shape (chunk_size, a, b)
        # 1) multiply each Jᵢ by C:  J_block_C[i,a,c] = ∑_b J[i,a,b] · prior_cov[b,c]
        J_block_C = jnp.tensordot(block, prior_covariance, axes=((2,), (0,)))
        #    shape → (chunk_size, a, b)
        # 2) contract with J again:
        #    partial[a,d] = ∑_{i,c} J_block_C[i,a,c] · block[i,d,c]
        partial = jnp.tensordot(J_block_C, block, axes=((0, 2), (0, 2)))
        #    shape → (a, a)
        return acc + partial, None

    # initialize accumulator (a×a) and scan over chunks
    A_sum, _ = jax.lax.scan(scan_fn, jnp.zeros((a, a), J_samples.dtype), Jc)
    # divide out the sample count and noise variance
    return(noise_precision / n) * A_sum 


def information_theoretic_dimension_reduction(
    key: Any,
    J_samples: jnp.ndarray,
    noise_precision: float,
    prior_precision: jnp.ndarray,
    max_input_dimension: int = None,
    max_output_dimension: int = None,
    prior_covariance: jnp.ndarray = None,
):
    """Document that this is for both input/output dimension reduction"""
    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    if max_input_dimension is not None:
        start = time. perf_counter()
        input_encodec_dict = estimate_input_active_subspace(
            key=key,
            J_samples=J_samples,
            noise_precision=noise_precision,
            prior_precision=prior_precision,
            subspace_rank=max_input_dimension,
        )
        input_encodec_dict["computation_time"] = time. perf_counter() - start
        print("Input subspace computation time:", input_encodec_dict["computation_time"])
    else:
        input_encodec_dict = dict()
    if max_output_dimension is not None:
        start = time. perf_counter()
        output_encodec_dict = estimate_output_informative_subspace(
            key=key,
            J_samples=J_samples,
            noise_precision=noise_precision,
            subspace_rank=max_output_dimension,
            prior_precision=prior_precision,
            prior_covariance=prior_covariance,
        )
        output_encodec_dict["computation_time"] = time. perf_counter() - start
        print("Output subspace computation time:", output_encodec_dict["computation_time"])
    else:
        output_encodec_dict = dict()
    return {"input": input_encodec_dict, "output": output_encodec_dict}


def moment_based_dimension_reduction(
    key: Any,
    input_covariance_matrix: jnp.ndarray = None,
    L2_inner_product_matrix: jnp.ndarray = None,
    output_samples: jnp.ndarray = None,
    max_input_dimension: int = None,
    max_output_dimension: int = None,
):
    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    if max_input_dimension is not None:
        start = time. perf_counter()
        input_encodec_dict = estimate_input_Karhunen_Loeve_subspace(
            key=key,
            input_covariance_matrix=input_covariance_matrix,
            L2_inner_product_matrix=L2_inner_product_matrix,
            subspace_rank=max_input_dimension,
        )
        input_encodec_dict["computation_time"] = time. perf_counter() - start
    else:
        input_encodec_dict = dict()
    if max_output_dimension is not None:
        start = time. perf_counter()
        output_encodec_dict = estimate_output_Proper_Orthogonal_Decomposition_subspace(
            key=key,
            output_samples=output_samples,
            subspace_rank=max_output_dimension,
        )
        output_encodec_dict["computation_time"] = time. perf_counter() - start
    else:
        output_encodec_dict = dict()
    return {"input": input_encodec_dict, "output": output_encodec_dict}


def estimate_output_Proper_Orthogonal_Decomposition_subspace(
    key: Any,
    output_samples,
    subspace_rank: int,
) -> Dict[str, jnp.ndarray]:
    """
    Compute the rank-r Proper Orthogonal Decomposition of output random variables
    and produce subspace encoder and decoder matrices.

    This is equivalent to producing the eigenpairs of a truncated SVD of output samples.

    """
    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    # SVD of output samples
    # return output_encodec_dict
    N = output_samples.shape[0]
    C = jnp.einsum("nd,ne->de", output_samples, output_samples) / (N - 1)
    computed_eigvals, computed_evecs = double_pass_randomized_eigh(
        key, C, subspace_rank, subspace_rank + 15, power_iters=2
    )
    return {
        "eigenvalues": computed_eigvals,
        "decoder": computed_evecs,
        "encoder": computed_evecs,
    }


def estimate_input_Karhunen_Loeve_subspace(
    key: Any,
    input_covariance_matrix: jnp.ndarray,
    L2_inner_product_matrix: jnp.ndarray,
    subspace_rank: int,
) -> Dict[str, jnp.ndarray]:
    """ "
    Compute the rank-r Karhunen Loeve (KL) expansion truncation of a Gaussian random variable
    and produce subspace encoder and decoder matrices.

    This function is applicable for finding the dominant eigenfunctions of a random variable
    distributed according to a Gaussian prior N(0, Cov), where Cov = input_covariance_matrix.
    It assumes the usage of a matrix (often a  Mass Matrix) used to define the L2 inner product,
    abbr as M. Thus, truncation of the KL expansion amounts to solving for the dominant eigenvectors
    x^i/values v^i such that Cov x^i = lambda^i M x^i. We do this via a double-pass randomized
    generalized eigendecomposition. The decoder is the eigenvectors, while the encoder is given by
    M @ eigenvectors.

    Parameters:
      key: Random key for the randomized eigendecomposition.
      input_covariance_matrix: An 2D dense covariance matrix array.
      L2_inner_product_matrix: A 2D matrix representing the L2 inner product.
      r: Target rank for the eigendecomposition.

    Returns:
      Dict with:
        "eigenvalues": 1D array of computed eigenvalues.
        "decoder": 2D array of computed eigenvectors.
        "encoder": 2D array computed as prior_precision @ eigenvectors.
    """
    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    computed_eigvals, computed_evecs = double_pass_randomized_gen_eigh(
        key,
        input_covariance_matrix,
        L2_inner_product_matrix,
        subspace_rank,
        subspace_rank + 5,
        power_iters=5,
        reorthog_iter=3,
    )
    return {
        "eigenvalues": computed_eigvals,
        "decoder": computed_evecs,
        "encoder": L2_inner_product_matrix @ computed_evecs,
    }


def estimate_input_active_subspace(
    key: Any,
    J_samples: jnp.ndarray,
    noise_precision: float,
    prior_precision: jnp.ndarray,
    subspace_rank: int,
) -> Dict[str, jnp.ndarray]:
    """
    Compute the rank-r active subspace encoder and decoder matrices.

    This function is applicable for finding the active subspace of the
    parameter under a Gaussian prior and additive Gaussian noise.
    It computes the matrix A as the mean of:
      J_samples[i]^T * (1/noise_variance) * J_samples[i]
    over all samples, and then uses a double-pass randomized generalized
    eigendecomposition to obtain the eigenvalues and eigenvectors. The
    decoder is the eigenvectors, while the encoder is given by
    prior_precision @ eigenvectors.

    Parameters:
      key: Random key for the randomized eigendecomposition.
      J_samples: 3D array of Jacobian samples with shape (N, a, b), where N
                is the number of samples.
      noise_precision: a positive scalar; noise precision is
                      1/noise_variance.
      prior_precision: 2D prior precision matrix.
      r: Target rank for the eigendecomposition.

    Returns:
      Dict with:
        "eigenvalues": 1D array of computed eigenvalues.
        "decoder": 2D array of computed eigenvectors.
        "encoder": 2D array computed as prior_precision @ eigenvectors.
    """
    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
        assert (
            J_samples.shape[0] % 10 == 0
        ), f"Number of samples ({J_samples.shape[0]}) must be divisible by chunk_size = 10"
    A = average_Jtranspose_sigmainv_J_chunked(J_samples, noise_precision, chunk_size=10)

    computed_eigvals, computed_evecs = double_pass_randomized_gen_eigh(
        key,
        A,
        prior_precision,
        subspace_rank,
        subspace_rank + 10,
        power_iters=1,
        reorthog_iter=3,
    )
    return {
        "eigenvalues": computed_eigvals,
        "decoder": computed_evecs,
        "encoder":  prior_precision @computed_evecs,
    }


def estimate_output_informative_subspace(
    key: Any,
    J_samples: jnp.ndarray,
    noise_precision: float,
    subspace_rank: int,
    prior_covariance: jnp.ndarray = None,
    prior_precision: jnp.ndarray = None,
) -> Dict[str, jnp.ndarray]:
    """
    Compute the rank-r encoder and decoder for informative data.

    This function is used for computing the informative data subspace
    under a Gaussian prior and additive Gaussian noise. It forms the
    matrix A as the mean over samples:
      A = (1/N) * sum(J_samples[i] * (prior_covariance/noise_variance) *
                      J_samples[i]^T)
    A double-pass randomized eigendecomposition is applied to obtain the
    eigenvalues and eigenvectors. The decoder scales the eigenvectors by
    sqrt(noise_variance), while the encoder scales them by the inverse of
    sqrt(noise_variance).

    Parameters:
      key: Random key for the eigendecomposition.
      J_samples: 3D array of Jacobian samples with shape (N, a, b), where N
                is the number of samples.
      noise_variance: positive scalar.
      r: Target rank for the eigendecomposition.
      prior_covariance: 2D prior covariance matrix.
      prior_precision: 2D prior precision matrix. If prior covariance is not
        provided, the prior precision will be used to perform a solve

    Returns:
      Dict with:
        "eigenvalues": 1D array of computed eigenvalues.
        "decoder": 2D array (scaled eigenvectors).
        "encoder": 2D array (inversely scaled eigenvectors).
    """
    if isinstance(key, int):
        key = jax.random.PRNGKey(key)
    if prior_covariance is None:
        A = average_JCJtranspose(J_samples, prior_precision)*noise_precision
    else:
        A = average_JCoversigmaJtranpose_chunked(J_samples, prior_covariance, noise_precision, chunk_size=10)

        # A = jnp.einsum(
        #     "iab,bc,idc->ad",
        #     J_samples,
        #     prior_covariance / noise_variance,
        #     J_samples / J_samples.shape[0],
        # )

    computed_eigvals, computed_evecs = double_pass_randomized_eigh(
        key, A, subspace_rank, p=subspace_rank + 15, power_iters=2
    )
    one_over_noise_sigma = jnp.sqrt(noise_precision)
    return {
        "eigenvalues": computed_eigvals,
        "decoder": computed_evecs/one_over_noise_sigma,
        "encoder": one_over_noise_sigma* computed_evecs,
    }



import time as gentime
from abc import ABC, abstractmethod
from typing import Dict, Optional, Tuple

import numpy as np
from numpy.typing import ArrayLike


class InputOuputAndDerivativesSampler(ABC):
    """
    Interface for functions and their derivatives for convenient training data
    sampling. This class is meant for objects that work solely with NumPy arrays.
    The input and output dimensions are fixed (provided in the constructor) and
    used for sampling training data.

    Attributes:
        input_dimension: The dimension of the input vector.
        output_dimension: The dimension of the output vector.
        output_computation_time: Optional time to compute outputs.
        jacobian_product_computation_time: Optional time to evaluate jacobian products.
    """

    @property
    def input_dimension(self) -> int:
        """
        The dimension of the input vector.

        Returns:
            An integer representing the input dimension.
        """
        return self._input_dimension

    @property
    def output_dimension(self) -> int:
        """
        The dimension of the output vector.

        Returns:
            An integer representing the output dimension.
        """
        return self._output_dimension

    @abstractmethod
    def sample_input(self) -> np.ndarray:
        """
        Sample an input vector from the function's domain. This method should
        return a NumPy array of shape (input_dimension,).

        Returns:
            A NumPy array representing a single input sample.
        """
        ...

    def init_matrix_jacobian_prod(self, *, matrix: Optional[np.ndarray] = None) -> None:
        """
        Set the matrix used for the matrix–Jacobian product computation.

        Parameters:
            matrix: A NumPy array used to modify the Jacobian product, or None.
        """
        self._matrix_jacobian_prod_matrix = matrix
        self.jacobian_product_computation_time = 0
        self._init_matrix_jacobian_prod(matrix=self._matrix_jacobian_prod_matrix)

    def init_value(self) -> None:
        self.output_computation_time = 0
        self._init_value()

    @abstractmethod
    def _init_value(self) -> None: ...

    @abstractmethod
    def _init_matrix_jacobian_prod(self, *, matrix: Optional[np.ndarray] = None) -> None: ...

    def value(self, input_sample: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the function value at a given
        input sample.

        Parameters:
            input_sample: A NumPy array of shape (input_dimension,).

        Returns:
            - A NumPy array of shape (output_dimension,) representing the
                function value.
        """
        start = gentime.perf_counter()
        value = self._value(input_sample)
        self.output_computation_time += gentime.perf_counter() - start
        return value

    @abstractmethod
    def _value(self, input_sample: np.ndarray) -> Tuple[np.ndarray, np.ndarray]: ...

    @abstractmethod
    def value_and_matrix_jacobian_prod(self, input_sample: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Compute the function value and the matrix–Jacobian product at a given
        input sample. The computation uses the matrix set by
        set_matrix_jacobian_prod.

        Parameters:
            input_sample: A NumPy array of shape (input_dimension,).

        Returns:
            A tuple containing:
              - A NumPy array of shape (output_dimension,) representing the
                function value.
              - A NumPy array representing the matrix–Jacobian product. Its shape
                depends on the applied matrix.
        """
        ...

    def extract_and_clear_computation_times(self) -> Dict[str, Optional[float]]:
        """
        Extract computation times (if they exist) and reset them to zero.

        This method looks for the attributes:
          - 'output_computation_time'
          - 'jacobian_product_computation_time'

        These attributes should have been used to count the total computation times
        for evaluating the function and its matrix jacobian product respectively

        Returns:
            A dictionary with keys:
              'output_computation_time' and
              'jacobian_product_computation_time' containing their values.
              If an attribute is not set, its value will be None.
        """
        times: Dict[str, Optional[float]] = {}
        if hasattr(self, "output_computation_time"):
            times["output_computation_time"] = self.output_computation_time
            self.output_computation_time = 0.0
        else:
            times["output_computation_time"] = None

        if hasattr(self, "jacobian_product_computation_time"):
            times["jacobian_product_computation_time"] = self.jacobian_product_computation_time
            self.jacobian_product_computation_time = 0.0
        else:
            times["jacobian_product_computation_time"] = None

        return times


class GaussianInputOuputAndDerivativesSampler(InputOuputAndDerivativesSampler):

    @property
    def noise_precision(self) -> ArrayLike:
        return self._noise_precision
    
    @property
    def precision(self) -> ArrayLike:
        return self._precision

    @property
    def L2_inner_product_matrix(self) -> ArrayLike:
        return self._L2_inner_product_matrix


def generate_full_Jacobian_data(
    *,
    sampler_wrapper: InputOuputAndDerivativesSampler,
    N_samples: int,
) -> Dict[str, np.ndarray]:
    """
    Generate full Jacobian data for dimension reduction.

    This is a thin wrapper around generate_reduced_training_data, which only
    requires an object that accepts NumPy arrays as input and returns NumPy arrays.

    Parameters:
        sampler_wrapper: An object implementing InputOuputAndDerivativesSampler.
        N_samples: The number of samples to generate.

    Returns:
        A dictionary containing training data and optional computation times.
    """
    return generate_reduced_training_data(sampler_wrapper=sampler_wrapper, N_samples=N_samples, generate_Jacobians=True)


def generate_output_data(*, sampler_wrapper: InputOuputAndDerivativesSampler, N_samples: int) -> Dict[str, np.ndarray]:
    """
    Generate full Jacobian data for dimension reduction.

    This is a thin wrapper around generate_reduced_training_data, which only
    requires an object that accepts NumPy arrays as input and returns NumPy arrays.

    Parameters:
        sampler_wrapper: An object implementing InputOuputAndDerivativesSampler.
        N_samples: The number of samples to generate.

    Returns:
        A dictionary containing training data and optional computation times.
    """
    return generate_reduced_training_data(
        sampler_wrapper=sampler_wrapper,
        N_samples=N_samples,
        generate_Jacobians=False,
    )


def generate_reduced_training_data(
    *,
    sampler_wrapper: InputOuputAndDerivativesSampler,
    N_samples: int,
    output_encoder: Optional[np.ndarray] = None,
    input_decoder: Optional[np.ndarray] = None,
    input_encoder: Optional[np.ndarray] = None,
    generate_Jacobians: bool = True,
    reduce_input_before: bool = False,
) -> Dict[str, np.ndarray]:
    """
    Generate reduced training data using a function and its derivatives.

    This function repeatedly samples inputs and computes the corresponding
    function outputs and matrix–Jacobian products. It then applies optional
    encoding/decoding to reduce the dimensionality of the inputs, outputs, and
    Jacobian products. This function requires that the provided sampler_wrapper works
    entirely with NumPy arrays.

    Parameters:
        sampler_wrapper: An object implementing InputOuputAndDerivativesSampler that works
            with NumPy arrays.
        N_samples: The number of samples to generate.
        output_encoder: Optional; a NumPy array of shape
            (sampler_wrapper.output_dimension, reduced_out_dim) to encode outputs.
        input_decoder: Optional; a NumPy array of shape
            (sampler_wrapper.input_dimension, reduced_in_dim) to reduce the input
            dimension of the Jacobian product.
        input_encoder: Optional; a NumPy array of shape
            (sampler_wrapper.input_dimension, reduced_in_dim) to encode inputs.

    Returns:
        A dictionary with keys:
          'inputs': A NumPy array of shape (N_samples, D_in_encoded), where
              D_in_encoded equals reduced_in_dim if input_encoder is provided,
              otherwise sampler_wrapper.input_dimension.
          'outputs': A NumPy array of shape (N_samples, D_out_encoded), where
              D_out_encoded equals reduced_out_dim if output_encoder is provided,
              otherwise sampler_wrapper.output_dimension.
          'Jacobians': A NumPy array whose shape depends on the provided
              encoders/decoders:
                - (N_samples, reduced_out_dim, reduced_in_dim) if both
                  output_encoder and input_decoder are provided.
                - (N_samples, reduced_out_dim, sampler_wrapper.input_dimension) if only
                  output_encoder is provided.
                - (N_samples, sampler_wrapper.output_dimension, reduced_in_dim) if only
                  input_decoder is provided.
                - (N_samples, sampler_wrapper.output_dimension,
                  sampler_wrapper.input_dimension) if neither is provided.
          'output_computation_time': The time for computing outputs (if set).
          'Jacobian_computation_time': The time for computing the Jacobian product
              (if set).
    """
    encoded_input_dimension = input_encoder.shape[1] if input_encoder is not None else sampler_wrapper.input_dimension
    encoded_output_dimension = (
        output_encoder.shape[1] if output_encoder is not None else sampler_wrapper.output_dimension
    )
    encoded_inputs = np.empty((N_samples, encoded_input_dimension))
    encoded_outputs = np.empty((N_samples, encoded_output_dimension))
    sampler_wrapper.init_value()
    if generate_Jacobians:
        encoded_jacobian_prod = np.empty((N_samples, encoded_output_dimension, encoded_input_dimension))
        sampler_wrapper.init_matrix_jacobian_prod(
            matrix=output_encoder if output_encoder is not None else np.identity(encoded_output_dimension)
        )

    print("Sampling...", flush=True)
    input_encoding_time = 0.0
    output_encoding_time = 0.0
    jacobian_decoding_time = 0.0
    for i in range(N_samples):
        input_sample = sampler_wrapper.sample_input()
        if input_encoder is not None:
            start = gentime.perf_counter()
            encoded_inputs[i] = input_sample @ input_encoder
            input_encoding_time += gentime.perf_counter() - start

        else:
            if reduce_input_before:
                encoded_inputs[i] = input_sample @ input_encoder
                input_sample = encoded_inputs[i] @ input_decoder.T
            else:
                encoded_inputs[i] = input_sample
        if generate_Jacobians:
            output_i, matrix_jacobian_prod_i = sampler_wrapper.value_and_matrix_jacobian_prod(input_sample)
        else:
            output_i = sampler_wrapper.value(input_sample)
        if output_encoder is not None:
            start = gentime.perf_counter()
            encoded_outputs[i] = output_i @ output_encoder
            output_encoding_time += gentime.perf_counter() - start
        else:
            encoded_outputs[i] = output_i
        if generate_Jacobians:
            if input_decoder is not None:
                start = gentime.perf_counter()
                encoded_jacobian_prod[i] = matrix_jacobian_prod_i @ input_decoder
                jacobian_decoding_time += gentime.perf_counter() - start
            else:
                encoded_jacobian_prod[i] = matrix_jacobian_prod_i
        print(f"{i+1}/{N_samples} samples generated.", flush=True)
    computation_times = sampler_wrapper.extract_and_clear_computation_times()
    input_key = "encoded_inputs" if input_encoder is not None else "inputs"
    output_key = "encoded_outputs" if output_encoder is not None else "outputs"
    jacobian_key = "encoded_Jacobians" if (input_decoder is not None or output_encoder is not None) else "Jacobians"

    results = {
        input_key: encoded_inputs,
        output_key: encoded_outputs,
        "output_computation_time": computation_times.get("output_computation_time"),
    }
    if generate_Jacobians:
        results[jacobian_key] = encoded_jacobian_prod
        results["Jacobian_encoding_time"] = computation_times.get("jacobian_product_computation_time")
        if input_decoder is not None:
            results["Jacobian_decoding_time"] = jacobian_decoding_time
    if input_encoder is not None:
        results["input_encoding_time"] = input_encoding_time
    if output_encoder is not None:
        results["output_encoding_time"] = output_encoding_time
    return results
